from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeVar, Generic, List

import numpy as np

from .file import ImageFile


class NodeOptions(ABC):
    pass


T = TypeVar('T', bound=NodeOptions, covariant=True)


class Node(ABC, Generic[T]):
    def __init__(self, options: T):
        self.options = options

    def update_options(self, options: T):
        self.options = options

    @abstractmethod
    def process(self, files: List[ImageFile]) -> List[ImageFile]:
        raise NotImplementedError

    @abstractmethod
    def single_process(self, file: ImageFile) -> ImageFile:
        raise NotImplementedError

    @abstractmethod
    def video_process(self, file: np.ndarray) -> np.ndarray:
        raise NotImplementedError



from dataclasses import dataclass
from typing import List, Optional

import numpy as np
from pepeline import resize
from reline.static import Node, NodeOptions, ImageFile
from .filter_type import FILTER_MAP, FilterType


@dataclass(frozen=True)
class ResizeOptions(NodeOptions):
    height: Optional[int] = None
    width: Optional[int] = None
    percent: Optional[float] = None
    filter: Optional[FilterType] = 'cubic_catrom'
    gamma_correction: Optional[bool] = False  # Потом уберу
    spread: Optional[bool] = False
    spread_size: Optional[int] = 2800


class ResizeNode(Node[ResizeOptions]):
    def __init__(self, options: ResizeOptions):
        super().__init__(options)

        if self.options.width is None and self.options.height is None and self.options.percent is None:
            raise ValueError('At least one of width, height, or percent must be specified.')

        self.filter = FILTER_MAP[self.options.filter]
        if options.percent:
            self.calculate_size = self._calculate_size__percent
        else:
            if self.options.width and not self.options.height:
                self.calculate_size = self._calculate_size__width
            elif self.options.height and not self.options.width:
                self.calculate_size = self._calculate_size__height
            else:
                self.calculate_size = self._calculate_size__sides

    def _calculate_size__height(self, height: int, width: int):
        return self.options.height, int(width * (self.options.height / height))

    def _calculate_size__width(self, height: int, width: int):
        result_width = self.options.width
        if self.options.spread and width > height:
            result_width = self.options.spread_size
        return int(height * (result_width / width)), result_width

    def _calculate_size__percent(self, height: int, width: int):
        return int(height * self.options.percent), int(width * self.options.percent)

    def _calculate_size__sides(self, _height: int, _width: int):
        return self.options.height, self.options.width

    def process(self, files: List[ImageFile]) -> List[ImageFile]:
        for file in files:
            h, w = self.calculate_size(*file.data.shape[:2])
            file.data = resize(file.data, h, w, self.filter).squeeze().clip(0, 1)

        return files

    def single_process(self, file: ImageFile) -> ImageFile:
        h, w = self.calculate_size(*file.data.shape[:2])
        file.data = resize(file.data, h, w, self.filter).squeeze().clip(0, 1)
        return file

    def video_process(self, file: np.ndarray) -> np.ndarray:
        h, w = self.calculate_size(*file.data.shape[:2])
        file = resize(file, h, w, self.filter).squeeze().clip(0, 1)
        return file



from dataclasses import dataclass
from typing import List, Optional, Literal

import numpy as np
from pepeline import DotType

from ._halftone_func import MODE_MAP, Mode
from reline.static import Node, NodeOptions, ImageFile
from ..resize.filter_type import FilterType, FILTER_MAP

DOT_TYPE_MAP = {
    'circle': DotType.CIRCLE,
    'cross': DotType.CROSS,
    'ellipse': DotType.ELLIPSE,
    'invline': DotType.INVLINE,
    'line': DotType.LINE,
}

TypeDot = Literal['line', 'cross', 'ellipse', 'invline', 'circle']


def _int_to_list(int_value: int | list[int]):
    if isinstance(int_value, int):
        return [int_value]
    return int_value


@dataclass(frozen=True)
class HalftoneOptions(NodeOptions):
    dot_size: Optional[int] | Optional[list[int]] = 7
    angle: Optional[int] | Optional[list[int]] = 0
    dot_type: Optional[TypeDot] | Optional[list[TypeDot]] = 'circle'
    halftone_mode: Optional[Mode] = 'gray'
    ssaa_scale: Optional[float] = None
    ssaa_filter: Optional[FilterType] = 'shamming4'


class HalftoneNode(Node[HalftoneOptions]):
    def __init__(self, options):
        super().__init__(options)
        self.dot_size = _int_to_list(options.dot_size)
        self.angle = _int_to_list(options.angle)
        self.halftone = MODE_MAP[options.halftone_mode]
        if isinstance(options.dot_type, str):
            self.dot_type = [DOT_TYPE_MAP[options.dot_type]]
        else:
            self.dot_type = [DOT_TYPE_MAP[dot_type] for dot_type in options.dot_type]
        self.scale = options.ssaa_scale
        self.ssaa_filter = FILTER_MAP[options.ssaa_filter]

    def process(self, files: List[ImageFile]) -> List[ImageFile]:
        for file in files:
            file.data = self.halftone(file.data.squeeze(), self.dot_size, self.angle, self.dot_type, self.scale, self.ssaa_filter)
        return files

    def single_process(self, file: ImageFile) -> ImageFile:
        file.data = self.halftone(file.data.squeeze(), self.dot_size, self.angle, self.dot_type, self.scale, self.ssaa_filter)
        return file

    def video_process(self, file: np.ndarray) -> np.ndarray:
        file = self.halftone(file.squeeze(), self.dot_size, self.angle, self.dot_type, self.scale, self.ssaa_filter)
        return file



from .file_reader import FileReaderNode, FileReaderOptions
from .file_writer import FileWriterNode, FileWriterOptions
from .folder_reader import FolderReaderNode, FolderReaderOptions
from .folder_writer import FolderWriterNode, FolderWriterOptions, FileFormat
from .resize import ResizeNode, ResizeOptions
from .upscale import UpscaleNode, UpscaleOptions
from .level import LevelNode, LevelOptions
from .halftone import HalftoneNode, HalftoneOptions
from .sharp import SharpNode, SharpOptions
from .cvt_color import CvtColorNode, CvtColorOptions
from .registry import Registry

INTERNAL_REGISTRY = (
    Registry()
    .set('resize', ResizeNode, ResizeOptions)
    .set('file_reader', FileReaderNode, FileReaderOptions)
    .set('file_writer', FileWriterNode, FileWriterOptions)
    .set('upscale', UpscaleNode, UpscaleOptions)
    .set('folder_reader', FolderReaderNode, FolderReaderOptions)
    .set('folder_writer', FolderWriterNode, FolderWriterOptions)
    .set('level', LevelNode, LevelOptions)
    .set('halftone', HalftoneNode, HalftoneOptions)
    .set('sharp', SharpNode, SharpOptions)
    .set('cvt_color', CvtColorNode, CvtColorOptions)
)

__all__ = [
    'FileReaderNode',
    'FileWriterNode',
    'FileReaderOptions',
    'FileWriterOptions',
    'FolderReaderNode',
    'FolderReaderOptions',
    'FolderWriterNode',
    'FolderWriterOptions',
    'FileFormat',
    'ResizeNode',
    'ResizeOptions',
    'UpscaleNode',
    'UpscaleOptions',
    'LevelNode',
    'LevelOptions',
    'HalftoneNode',
    'HalftoneOptions',
    'SharpNode',
    'SharpOptions',
    'CvtColorNode',
    'CvtColorOptions',
    'INTERNAL_REGISTRY',
]



import traceback
from typing import Annotated, Optional
import orjson
import typer
from . import Pipeline, __version__
from rich import print
from rich.table import Table

app = typer.Typer(pretty_exceptions_show_locals=False)


def version_callback(value: bool):
    if not value:
        return

    table = Table()
    table.add_column('Dependency', style='cyan', no_wrap=True)
    table.add_column('Version', style='magenta')
    table.add_column('Details', style='green')

    table.add_row('reline', __version__ or 'unknown', '-')

    import resselt

    table.add_row('resselt', getattr(resselt, '__version__', 'unknown'), '-')

    import torch

    cuda_status = 'Yes' if torch.cuda.is_available() else 'No'
    table.add_row('torch', torch.__version__, f'CUDA Available: {cuda_status}')

    print(table)
    raise typer.Exit()


@app.command()
def main(
    config: str = typer.Option('config.json', '--config', '-c', help='Path to the configuration JSON file'),
    version: Annotated[Optional[bool], typer.Option('--version', callback=version_callback)] = None,
):
    try:
        with open(config, encoding='utf-8') as f:
            data = f.read()
            data = orjson.loads(data)

        Pipeline.from_json(data).process_linear()
    except orjson.JSONDecodeError:
        typer.echo(f"Error: Invalid JSON in config file '{config}': {traceback.format_exc()}")
        raise typer.Exit(1)


def cli():
    """Entry point for the CLI."""
    app()


if __name__ == '__main__':
    cli()



from __future__ import annotations

import os.path
from dataclasses import dataclass
from typing import List, Optional, Literal
import logging

from pepeline import read, ImgFormat, ImgColor

from reline.static import Node, NodeOptions, ImageFile

MODE_MAP = {'rgb': ImgColor.RGB, 'gray': ImgColor.GRAY, 'dynamic': ImgColor.DYNAMIC}

Mode = Literal['rgb', 'gray', 'dynamic']


class ImageIterator:
    def __init__(self, image_paths: list, dir_path: str, mode: ImgColor):
        self.current = 0
        self.image_paths = image_paths
        self.dir_path = dir_path
        self.end = len(image_paths)
        self.mode = mode

    def __len__(self):
        return self.end

    def __iter__(self):
        return self

    def __next__(self):
        if self.current >= self.end:
            raise StopIteration
        else:
            try:
                file_path = self.image_paths[self.current]
                commonprefix = os.path.commonprefix([self.dir_path, file_path])
                dirpath = os.path.dirname(os.path.relpath(file_path, commonprefix))
                basename, _ = os.path.splitext(os.path.basename(file_path))
                data = read(file_path, self.mode, ImgFormat.F32)
                file = ImageFile(data, basename, dirpath)
                self.current += 1
                return file
            except Exception as e:
                logging.warning(f'image {basename} not decoded due to error: {e}')
                return None


@dataclass(frozen=True)
class FolderReaderOptions(NodeOptions):
    path: str
    recursive: Optional[bool] = False
    mode: Optional[Mode] = 'dynamic'


class FolderReaderNode(Node[FolderReaderOptions]):
    def __init__(self, options: FolderReaderOptions):
        super().__init__(options)
        self.mode = MODE_MAP[options.mode]
        self.dir_path = os.path.abspath(self.options.path)

    def _scandir(self, dir_path: str):
        file_paths = []

        try:
            for entry in os.scandir(dir_path):
                if entry.is_file():
                    file_paths.append(os.path.abspath(entry.path))
                elif entry.is_dir() and self.options.recursive:
                    file_paths.extend(self._scandir(os.path.abspath(entry.path)))
        except OSError as e:
            logging.error(f'Error scanning directory {dir_path}: {e}')

        return file_paths

    def process(self, _) -> List[ImageFile]:
        file_paths = self._scandir(self.dir_path)
        files = []
        basename = None
        for file_path in file_paths:
            try:
                commonprefix = os.path.commonprefix([self.dir_path, file_path])
                dirpath = os.path.dirname(os.path.relpath(file_path, commonprefix))
                basename, _ = os.path.splitext(os.path.basename(file_path))

                data = read(file_path, self.mode, ImgFormat.F32)

                file = ImageFile(data, basename, dirpath)
                files.append(file)
            except Exception as e:
                logging.warning(f'image {basename} not decoded due to error: {e}')
                continue

        return files

    def single_process(self, _) -> ImageIterator:
        file_paths = self._scandir(self.dir_path)
        return ImageIterator(file_paths, self.dir_path, self.mode)

    def video_process(self, _):
        raise ValueError('Video scale does not support folder read')
